using HELLo_World.MapData;
using HELLo_World.MapData.Buildings;
using HELLo_World.MapData.People;
using HELLo_World.Model.StaticData;
using HELLo_World.Model.EventArguments;
using HELLo_World.Persistence;
using System;
using System.Threading.Tasks;
using HELLo_World.Model.ModelExceptions;
using HELLo_World.Model.Information;
using System.Collections.Generic;
using System.Linq;

namespace HELLo_World.Model
{
    /// <summary>
    /// HELLoWorld game model type
    /// </summary>

    public class GameModel
    {
        #region Constants

        public  const Int32 MapSize = 15;   //Size of game table
        private const Int32 CateringStartingFee = 10;   //Starting fee of catering services
        private const Int32 EntertainmentStartingFee = 10;  //Starting fee of entertainment services
        private const Int32 BuildTime = 20;     //Length of building
        private const Int32 RepairTime = 20;    //Length of repairing

        #endregion

        #region Fields

        private IGamePersistence _dataAccess;   //Data access
        private Building[,] _map;   //Game table
        private ParkStatus _parkStatus; //Status of the park
        private Int32 _buildingId = 1;  //Starting building id
        private Int32 _personId = 1;    //Starting person id
        private List<Int32> _roadIds;   //List of road ids
        private List<EntertainmentFacility> _entertainmentFacilities;   //List of entertainment facilities
        private List<CateringFacility> _cateringFacilities; //List of catering facilities
        private List<Int32> _cateringIds;   //List of catering ids
        private List<Int32> _entertainmentIds;  //List of entertainment ids
        private Speed _gameSpeed;   //Game speed

        #endregion

        #region Properties

        /// <summary>
        /// Gets/sets BuildingData BuildingData
        /// </summary>
        public BuildingData BuildingData { get; private set; }
        
        /// <summary>
        /// Gets/sets GameMap GameMap
        /// </summary>
        public GameMap GameMap { get; private set; }

        /// <summary>
        /// Gets/sets the value of the Speed _gameSpeed
        /// </summary>
        public Speed GameSpeed
        {
            get
            {
                return _gameSpeed;
            }
            set
            {
                if (_gameSpeed == value) return;
                _gameSpeed = value;
                OnSpeedChanged(value);
            }
        }

        /// <summary>
        /// Gets/sets the value of the ParkStatus _parkStatus
        /// </summary>
        public ParkStatus ParkStatus
        {
            get { return _parkStatus; }
            set
            {
                if (value != _parkStatus)
                {
                    _parkStatus = value;
                    if (value == ParkStatus.CLOSED)
                    {
                        SendHomeVisitors();
                    }
                    OnParkStatusChanged();
                }
            }
        }

        /// <summary>
        /// Gets the value of the Int32 GameMap.Repairmen.Count.
        /// Adds new repairman to GameMap.Repairmen
        /// </summary>
        public Int32 NumberOfRepairmen
        {
            get { return GameMap.Repairmen.Count; }
            set
            {
                if (value < 0) throw new ArgumentException();
                if (value == GameMap.Repairmen.Count) return;
                if (value > GameMap.Repairmen.Count)
                {
                    int curr = GameMap.Repairmen.Count;
                    for (int i = 0; i < value - curr; i++)
                    {
                        Repairman rep = new Repairman(-(++_personId), new Tuple<int, int>(GameMap.Buildings[0].Coord.Item1 + 1, GameMap.Buildings[0].Coord.Item2 + 2), new Tuple<int, int>(25, 25));
                        rep.CurrentActivity = Activity.ARRIVED_TO_ROAD;
                        GameMap.Repairmen.Add(rep);
                        OnPersonAdded(rep);
                    }
                }
                else
                {
                    int diff = GameMap.Repairmen.Count - value;
                    List<Repairman> del = new List<Repairman>();
                    foreach (var repairman in GameMap.Repairmen)
                    {
                        if (repairman.CurrentActivity != MapData.People.Activity.REPAIRING)
                        {
                            OnPersonDeleted(repairman);
                            del.Add(repairman);
                            diff--;
                            if (diff == 0)
                            {
                                foreach (Repairman rep in del)
                                {
                                    GameMap.Repairmen.Remove(rep);
                                }
                                return;
                            }
                        }
                    }
                    foreach (var repairman in GameMap.Repairmen)
                    {
                        repairman.CurrentBuilding.Status = FacilityStatus.FAULTY;
                        repairman.CurrentBuilding.UserList.Clear();
                        repairman.CurrentBuilding.TimeSpentInStatus = 0;
                        OnPersonDeleted(repairman);
                        del.Add(repairman);
                        diff--;
                        if (diff == 0)
                        {
                            foreach (Repairman rep in del)
                            {
                                GameMap.Repairmen.Remove(rep);
                            }
                            return;
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Gets/sets Boolean PopupIsActive
        /// </summary>
        public Boolean PopupIsActive { get; set; }

        /// <summary>
        /// Gets the value of the field's BuildingTypeId
        /// </summary>
        /// <param name="x">Column index</param>
        /// <param name="y">Row index</param>
        /// <returns>Field's BuildingTypeId</returns>
        public Int32 this[int x, int y]
        {
            get
            {
                if (x < 0 || x >= _map.GetLength(0))
                    throw new ArgumentException("Bad column index.", "x");
                if (y < 0 || y >= _map.GetLength(1))
                    throw new ArgumentException("Bad row index.", "y");

                if (_map[y, x] is null) return -1;
                return _map[y, x].BuildingTypeId;
            }
        }

        #endregion

        #region Constructor

        /// <summary>
        /// Instantiation of GameModel
        /// </summary>
        /// <param name="dataAccess">Data access</param>
        public GameModel(IGamePersistence dataAccess)
        {
            _dataAccess = dataAccess;
            GameSpeed = Speed.PAUSE;

            _entertainmentFacilities = new List<EntertainmentFacility>();
            _cateringFacilities = new List<CateringFacility>();
            BuildingData = new BuildingData();
            BuildingDataInitializer.Initialize(BuildingData);
        }

        #endregion

        #region Events
        
        /// <summary>
        /// Event for canvas refresh
        /// </summary>
        public event EventHandler VisitorsChanged;

        /// <summary>
        /// Event for person's data change
        /// </summary>
        public event EventHandler<PersonChangedEventArgs> PersonChanged;

        /// <summary>
        /// Event for person add to GameMap
        /// </summary>
        public event EventHandler<PersonChangedEventArgs> PersonAdded;

        /// <summary>
        /// Event for person delete from GameMap
        /// </summary>
        public event EventHandler<PersonChangedEventArgs> PersonDeleted;

        /// <summary>
        /// Event for building add to GameMap
        /// </summary>
        public event EventHandler<BuildingChangedEventArgs> BuildingAdded;

        /// <summary>
        /// Event for building delete from GameMap
        /// </summary>
        public event EventHandler<BuildingChangedEventArgs> BuildingDestroyed;

        /// <summary>
        /// Event for building's construction finished
        /// </summary>
        public event EventHandler<BuildingChangedEventArgs> BuildingChanged;

        /// <summary>
        /// Event for money change
        /// </summary>
        public event EventHandler MoneyChanged;

        /// <summary>
        /// Event for new game start
        /// </summary>
        public event EventHandler<NewGameEventArgs> NewGameStarted;

        /// <summary>
        /// Event for speed change
        /// </summary>
        public event EventHandler<SpeedChangedEventArgs> SpeedChanged;

        /// <summary>
        /// Event for tick
        /// </summary>
        public event EventHandler TimePassed;

        /// <summary>
        /// Event for part status change
        /// </summary>
        public event EventHandler ParkStatusChanged;

        /// <summary>
        /// Event for park close
        /// </summary>
        public event EventHandler EnableDestroy;

        /// <summary>
        /// Event for info refresh
        /// </summary>
        public event EventHandler PopupRefreshed;

        #endregion

        #region Public methods

        /// <summary>
        /// Starts a new game
        /// </summary>
        /// <param name="name">Park name</param>
        public void NewGame(String name)
        {
            _buildingId = 1;
            GameMap = new GameMap(name);
            NumberOfRepairmen = 0;
            _roadIds = new List<int>();
            _entertainmentIds = new List<int>();
            _cateringIds = new List<int>();
            _map = new Building[MapSize, MapSize];
            ParkStatus = ParkStatus.CLOSED;
            for(int i = 0; i < MapSize; ++i)
            {
                for(int j = 0; j < MapSize; ++j)
                {
                    _map[i, j] = null;
                }
            }
            Build(new Tuple<int, int>(13, 5), -1);
            OnNewGameStarted(GameMap, name);

            GameSpeed = Speed.PAUSE;
        }

        /// <summary>
        /// Adds new building to GameMap
        /// </summary>
        /// <param name="coord">Coordinates of the building's top left corner</param>
        /// <param name="buildingTypeId">Building type id</param>
        public void Build(Tuple<Int32, Int32> coord, Int32 buildingTypeId)
        {
            StaticBuildingData building = BuildingData.GetBuilding(buildingTypeId);

            if (building.Price > GameMap.Money) throw new BuildException(BuildErrorType.NOT_ENOUGH_MONEY);
            for (int i = coord.Item1; i < coord.Item1 + building.Size.Item1; i++)
            {
                for (int j = coord.Item2; j < coord.Item2 + building.Size.Item2; j++)
                {
                    if (i >= MapSize || j >= MapSize) throw new BuildException(BuildErrorType.OFF_MAP);
                    if (_map[i, j] != null)           throw new BuildException(BuildErrorType.SPACE_OCCUPIED);
                }
            }

            GameMap.Money -= building.Price;
            OnMoneyChanged();
            Building newBuilding = null;
            StaticBuildingData sBuild = null;
            if (buildingTypeId == -1)
            {
                newBuilding = new Gate(_buildingId++, buildingTypeId, coord);
                GameMap.Buildings.Add(newBuilding);
                for (int i = coord.Item1; i < coord.Item1 + building.Size.Item1; i++)
                {
                    for (int j = coord.Item2; j < coord.Item2 + building.Size.Item2; j++)
                    {
                        if(i == coord.Item1 && j == coord.Item2 + 2)
                        {
                            _map[i, j] = new Road(-1, 0, new Tuple<int, int>(coord.Item1, coord.Item2 + 2));
                        } else if(i == coord.Item1 + 1 && j == coord.Item2 + 2)
                        {
                            _map[i, j] = new Road(-2, 0, new Tuple<int, int>(coord.Item1 + 1, coord.Item2 + 2));
                        } else
                        {
                            _map[i, j] = newBuilding;
                        }
                    }
                }
            }
            else if (buildingTypeId == 0)
            {
                newBuilding = new Road(_buildingId++, buildingTypeId, coord);
                _roadIds.Add(newBuilding.BuildingId);
                GameMap.Buildings.Add(newBuilding);
                _map[coord.Item1, coord.Item2] = newBuilding;
            }
            else 
            {
                if (buildingTypeId > 0 && buildingTypeId < 100) 
                {
                    sBuild = BuildingData.GetBuilding(buildingTypeId);
                    CateringData fData = sBuild as CateringData;
                    newBuilding = new CateringFacility(_buildingId++, buildingTypeId, coord, CateringStartingFee);
                    _cateringIds.Add(newBuilding.BuildingId);
                    _cateringFacilities.Add(newBuilding as CateringFacility);
                    GameMap.Buildings.Add(newBuilding);
                }   
                else if (buildingTypeId > 99)
                {
                    sBuild = BuildingData.GetBuilding(buildingTypeId);
                    EntertainmentData fData = sBuild as EntertainmentData;
                    newBuilding = new EntertainmentFacility(_buildingId++, buildingTypeId, coord, EntertainmentStartingFee, fData.MinimumUsers);
                    _entertainmentIds.Add(newBuilding.BuildingId);
                    _entertainmentFacilities.Add(newBuilding as EntertainmentFacility);
                    GameMap.Buildings.Add(newBuilding);
                }     
                for (int i = coord.Item1; i < coord.Item1 + building.Size.Item1; i++)
                {
                    for (int j = coord.Item2; j < coord.Item2 + building.Size.Item2; j++)
                    {
                        _map[i, j] = newBuilding;
                    }
                }
            }

            OnBuildingAdded(newBuilding);
        }

        /// <summary>
        /// Destroys building from GameMap
        /// </summary>
        /// <param name="coord">Coordinates of the building's top left corner</param>
        /// <param name="size">Size of the building</param>
        public void Destroy(Tuple<Int32, Int32> coord, Tuple<Int32, Int32> size)
        {
            if (_map[coord.Item1, coord.Item2] == null) throw new DestroyException(DestroyErrorType.NO_BUILDING);
            Gate gate = GameMap.Buildings[0] as Gate;
            StaticBuildingData gateData = BuildingData.GetBuilding(gate.BuildingTypeId);
            if (coord.Item1 >= gate.Coord.Item1 && coord.Item1 < gate.Coord.Item1 + gateData.Size.Item1  && coord.Item2 >= gate.Coord.Item2 && coord.Item2 < gate.Coord.Item2 + gateData.Size.Item2)
                throw new DestroyException(DestroyErrorType.GATE);
            if (GameMap.Visitors.Count != 0) throw new DestroyException(DestroyErrorType.NOT_EMPTY);

            foreach (Building building in GameMap.Buildings)
            {
                if(building.Coord.Item1 == coord.Item1 && building.Coord.Item2 == coord.Item2)
                {
                    for(int i = coord.Item1; i < coord.Item1 + size.Item1; ++i)
                    {
                        for(int j = coord.Item2; j < coord.Item2 + size.Item2; ++j)
                        {
                            _map[i, j] = null;
                        }
                    }
                    if(building.BuildingTypeId != -1)
                    {
                        if(building is Road)
                        {
                            _roadIds.Remove(building.BuildingId);
                        } else if(building is CateringFacility)
                        {
                            _cateringIds.Remove(building.BuildingId);
                        } else
                        {
                            _entertainmentIds.Remove(building.BuildingId);
                        }
                        OnBuildingDestroyed(building);
                        GameMap.Buildings.Remove(building);
                        break;
                    }
                }
            }
        }

        /// <summary>
        /// Gets informations of a building
        /// </summary>
        /// <param name="buildingId">Unique building id</param>
        /// <returns>Informations of a building</returns>
        public BuildingInfo GetBuildingInfo(Int32 buildingId)
        {
            BuildingInfo bInfo = null;
            Building build = null;
            StaticBuildingData sBuild = null;

            foreach (Building building in GameMap.Buildings)
            {
                if (building.BuildingId == buildingId)
                {
                    build = building;
                    sBuild = BuildingData.GetBuilding(build.BuildingTypeId);
                    break;
                }
            }

            if (build != null && !(sBuild is RoadData) && !(sBuild is GateData))
            {
                Facility fac = build as Facility;
                FacilityData fData = sBuild as FacilityData;
                Tuple<Tuple<Int32, Int32>, Int32> cap = new Tuple<Tuple<int, int>, int>(new Tuple<int, int>(fac.UserList.Count, fac.VisitorQueue.Count), fData.Capacity);
                if(build is EntertainmentFacility)
                {
                    bInfo = new EntertainmentInfo(buildingId, cap, fac.Fee, fData.Name, fac.Status, (fac as EntertainmentFacility).MinimumCapacity, (fData as EntertainmentData).Adrenalin);
                }
                else
                {
                    bInfo = new BuildingInfo(buildingId, cap, fac.Fee, fData.Name, fac.Status);
                } 
            }
            return bInfo;
        }

        /// <summary>
        /// Gets informations of a visitor
        /// </summary>
        /// <param name="visitorId">Unique visitor id</param>
        /// <returns>Informaions of a visitor</returns>
        public VisitorInfo GetVisitorInfo(Int32 visitorId)
        {
            VisitorInfo vInfo = null;

            foreach (var visitor in GameMap.Visitors)
            {
                if (visitor.PersonId == visitorId)
                {
                    vInfo = new VisitorInfo(visitor.AdrenalinLevel, visitor.FoodLevel, visitor.HappinessLevel, visitor.Money, visitor.CurrentActivity, visitor.CurrentStatus);
                    break;
                }
            }
            return vInfo;
        }

        /// <summary>
        /// Execute actions on every tick
        /// </summary>
        public void AdvanceTime()
        {
            if (GameMap.Visitors.Count == 0 && ParkStatus == ParkStatus.CLOSED)
            {
                bool enableDestroy = true;
                foreach(var repairman in GameMap.Repairmen)
                {
                    enableDestroy = enableDestroy && (repairman.CurrentActivity == Activity.RESTING);
                }
                if(enableDestroy)
                {
                    GameSpeed = Speed.PAUSE;
                    OnEnableDestroy();
                }
            }

            List<Visitor> activeVisitors = new List<Visitor>();
            foreach (Visitor visitor in GameMap.Visitors)
            {

                if(visitor.CurrentActivity != Activity.ARRIVED_TO_EXIT)
                {
                    activeVisitors.Add(visitor);
                }

                visitor.CurrentBuilding = (_map[visitor.Coord.Item1, visitor.Coord.Item2] is Facility) ? (_map[visitor.Coord.Item1, visitor.Coord.Item2] as Facility) : null;
                switch(visitor.CurrentActivity)
                {
                    case Activity.WANDERING:
                        visitor.Step();
                        visitor.LevelsChange();
                        visitor.ChooseStatus();
                        OnPersonChanged(visitor);
                        break;

                    case Activity.MOVING:
                        visitor.Step();
                        visitor.LevelsChange();
                        if (visitor.CurrentActivity == Activity.WAITING)
                        {
                            if(visitor.Money - visitor.CurrentBuilding.Fee < 0 || visitor.CurrentBuilding.Status == FacilityStatus.FAULTY || visitor.CurrentBuilding.Status == FacilityStatus.UNDER_CONSTRUCTION || visitor.CurrentBuilding.Status == FacilityStatus.REPAIRING)
                            {
                                visitor.CurrentActivity = Activity.ARRIVED_TO_ROAD;
                            }
                            else
                            {
                                if (visitor.CurrentBuilding is CateringFacility)
                                {
                                    JoinCatering(visitor, (visitor.CurrentBuilding as CateringFacility));
                                }
                                else if (visitor.CurrentBuilding is EntertainmentFacility)
                                {
                                    JoinEntertainment(visitor, (visitor.CurrentBuilding as EntertainmentFacility));
                                }
                            }      
                        }
                        OnPersonChanged(visitor);
                        break;

                    case Activity.LEAVING:
                        visitor.Step();
                        OnPersonChanged(visitor);
                        break;

                    case Activity.ARRIVED_TO_ROAD:
                        Stack<Tuple<int,int>> path = null;
                        visitor.ChooseStatus();
                        if (visitor.CurrentStatus != Status.SATISFIED)
                        {
                            if(visitor.CurrentStatus == Status.SAD || visitor.CurrentStatus == Status.NOT_ENOUGH_MONEY)
                            {
                                visitor.CurrentActivity = Activity.LEAVING;
                            }
                            else
                            {
                                visitor.CurrentActivity = Activity.MOVING;

                            }
                            path = PathToDest(visitor.Coord, ChooseDest(visitor));
                            if (!(path is null))
                            {
                                visitor.Path = path;
                            }
                        }
                        if(path is null)
                        {
                            visitor.CurrentActivity = Activity.WANDERING;
                            path = PathToDest(visitor.Coord, ChooseDest(visitor));
                            visitor.Path = path;
                            if (path is null)
                            {
                                visitor.CurrentActivity = Activity.ARRIVED_TO_ROAD;
                            }
                        }
                        break;
                    case Activity.ARRIVED_TO_EXIT:
                        OnPersonDeleted(visitor);
                        break;
                    case Activity.WAITING:
                        visitor.LevelsChange();
                        if(visitor.CurrentStatus == Status.SAD)
                        {
                            visitor.CurrentActivity = Activity.ARRIVED_TO_ROAD;
                            if(visitor.CurrentBuilding != null)
                            {
                                visitor.CurrentBuilding.Leave(visitor);
                            }
                        }
                        break;
                    case Activity.USING:
                        visitor.CurrentInteractionTimeLeft--;
                        if (visitor.CurrentInteractionTimeLeft == 0)
                        {
                            visitor.CurrentActivity = Activity.ARRIVED_TO_ROAD;
                            if(visitor.CurrentBuilding is CateringFacility)
                            {
                                GameMap.Money += visitor.CurrentBuilding.Fee - (BuildingData.GetBuilding(visitor.CurrentBuilding.BuildingTypeId) as CateringData).UseMaintenance;
                            }
                            else
                            {
                                GameMap.Money += visitor.CurrentBuilding.Fee;
                            }
                        }
                        break;
                }
            }
            GameMap.Visitors = activeVisitors;

            GameMap.Time++;

            foreach (var building in GameMap.Buildings)
            {
                if (building is Facility)
                {
                    Facility fac = building as Facility;

                    if(GameMap.Time % 100 == 0)
                    {
                        GameMap.Money -= (BuildingData.GetBuilding(fac.BuildingTypeId) as FacilityData).RegularMaintenance;
                    }

                    
                    switch (fac.Status)
                    {
                        case FacilityStatus.OPERATING:
                            if (ParkStatus == ParkStatus.CLOSED)
                            {
                                (building as Facility).Empty();
                                fac.Status = FacilityStatus.WAITING;
                            }
                            else
                            {
                                if (fac is CateringFacility)
                                {
                                    RefreshCatering(fac as CateringFacility);
                                }
                                else
                                {
                                    RefreshEntertainment(fac as EntertainmentFacility);
                                    if (fac.Status == FacilityStatus.WAITING)
                                    {
                                        GameMap.Money -= (BuildingData.GetBuilding(fac.BuildingTypeId) as EntertainmentData).UseMaintenance;

                                    }
                                }
                            }
                            break;
                        case FacilityStatus.WAITING:
                            if (ParkStatus == ParkStatus.CLOSED)
                            {
                                (building as Facility).Empty();
                                fac.Status = FacilityStatus.WAITING;
                            }
                            else if (fac is EntertainmentFacility)
                            { 
                                RefreshEntertainment(fac as EntertainmentFacility);
                            }
                            break;
                        case FacilityStatus.UNDER_CONSTRUCTION:
                            fac.TimeSpentInStatus++;
                            if (fac.TimeSpentInStatus >= BuildTime)
                            {
                                fac.TimeSpentInStatus = 0;
                                fac.Status = FacilityStatus.WAITING;
                                OnBuildingChanged(building);
                            }
                            break;
                        case FacilityStatus.FAULTY:
                            if(!(building as EntertainmentFacility).RepairManOnTheWay)
                            {
                                foreach (var repairMan in GameMap.Repairmen)
                                {
                                    if (repairMan.CurrentActivity != Activity.REPAIRING && repairMan.CurrentActivity != Activity.MOVING && repairMan.CurrentActivity != Activity.WAITING)
                                    {
                                        repairMan.Path = null;
                                        repairMan.CurrentActivity = Activity.MOVING;
                                        repairMan.Path = PathToDest(repairMan.Coord, building.BuildingId);
                                        (building as EntertainmentFacility).RepairManOnTheWay = true;
                                        break;
                                    }
                                }
                            }
                            break;
                        case FacilityStatus.REPAIRING:
                            fac.TimeSpentInStatus++;
                            if (fac.TimeSpentInStatus >= RepairTime)
                            {
                                fac.TimeSpentInStatus = 0;
                                fac.Status = FacilityStatus.WAITING;
                                (fac as EntertainmentFacility).RepairManOnTheWay = false;
                            }
                            break;
                        default:
                            break;
                    }
                }

            }

            foreach (var repairMan in GameMap.Repairmen)
            {
                repairMan.CurrentBuilding = (_map[repairMan.Coord.Item1, repairMan.Coord.Item2] is Facility) ? (_map[repairMan.Coord.Item1, repairMan.Coord.Item2] as Facility) : null;
                switch (repairMan.CurrentActivity)
                {
                    case Activity.WANDERING:
                        repairMan.Step();
                        OnPersonChanged(repairMan);
                        break;

                    case Activity.MOVING:
                        repairMan.Step();
                        OnPersonChanged(repairMan);
                        break;
                    
                    case Activity.GOING_TO_REST:
                        repairMan.Step();
                        OnPersonChanged(repairMan);
                        break;

                    case Activity.ARRIVED_TO_ROAD:
                        Stack<Tuple<int, int>> path = null;
                        if (path is null)
                        {
                            if(ParkStatus == ParkStatus.CLOSED)
                            {
                                repairMan.Path  = PathToDest(repairMan.Coord, -2);
                                repairMan.CurrentActivity = Activity.GOING_TO_REST;
                            }
                            else
                            {
                                repairMan.CurrentActivity = Activity.WANDERING;
                                path = PathToDest(repairMan.Coord, ChooseDest(repairMan));
                                repairMan.Path = path;
                                if (path is null)
                                {
                                    repairMan.CurrentActivity = Activity.ARRIVED_TO_ROAD;
                                }
                            }
                        }
                        break;

                    case Activity.WAITING:
                        repairMan.CurrentBuilding.TimeSpentInStatus = 0;
                        repairMan.CurrentBuilding.Status = FacilityStatus.REPAIRING;
                        repairMan.CurrentActivity = Activity.REPAIRING;
                        break;

                    case Activity.REPAIRING:
                        if (repairMan.CurrentBuilding.Status == FacilityStatus.WAITING)
                        {
                            repairMan.CurrentActivity = Activity.ARRIVED_TO_ROAD;
                        }
                        break;
                    case Activity.RESTING:
                        if(ParkStatus == ParkStatus.OPEN)
                        {
                            repairMan.CurrentActivity = Activity.ARRIVED_TO_ROAD;
                        }
                        break;
                }
            }

            SpawnVisitors();
            
            OnTimePassed();
            OnVisitorsChanged();
            if(PopupIsActive)
            {
                OnPopupRefreshed();
            }
        }

        /// <summary>
        /// Sets the ticket price
        /// </summary>
        /// <param name="newTicket">New price of the ticket</param>
        public void SetTicket(Int32 newTicket)
        {
            if (newTicket < 0)
            {
                throw new ArgumentException();
            }
            GameMap.Ticket = newTicket;
        }

        /// <summary>
        /// Sets the fee of a building
        /// </summary>
        /// <param name="buildingId">Unique building id</param>
        /// <param name="newFee">New fee for a building</param>
        public void SetBuildingFee(Int32 buildingId, Int32 newFee)
        {
            if (newFee < 0) throw new ArgumentException();
            Facility fac = null;
            foreach (Building building in GameMap.Buildings)
            {
                if (building.BuildingId == buildingId && building is Facility)
                {
                    fac = building as Facility;
                    break;
                }
            }
            if (fac == null) throw new ArgumentException();
            fac.Fee = newFee;
        }

        /// <summary>
        /// Sets minimum capacity of an entertainment facility
        /// </summary>
        /// <param name="buildingId">Unique building id</param>
        /// <param name="cap">New minimum capacity of an entertainment facility</param>
        public void SetMinimumCapacity(Int32 buildingId, Int32 cap)
        {
            if (cap < 0) throw new ArgumentException();
            Facility fac = null;
            foreach (Building building in GameMap.Buildings)
            {
                if (building.BuildingId == buildingId && building is EntertainmentFacility)
                {
                    fac = building as EntertainmentFacility;
                    break;
                }
            }
            if (fac == null) throw new ArgumentException();
            (fac as EntertainmentFacility).MinimumCapacity = cap;
        }

        #endregion

        #region Private methods

        /// <summary>
        /// Adds visitors to GameMap
        /// </summary>
        private void SpawnVisitors()
        {
            if (ParkStatus == ParkStatus.CLOSED) return;
            Int32 maxVisitorCount = 0;
            Int32 maxAdrenalin = -1;
            Int32 minAdrenalin = 101;
            foreach (Facility facility in _cateringFacilities)
            {
                var facilityData = BuildingData.GetBuilding(facility.BuildingTypeId);
                maxVisitorCount += (facilityData as FacilityData).Capacity;
            }
            foreach (Facility facility in _entertainmentFacilities)
            {
                var facilityData = BuildingData.GetBuilding(facility.BuildingTypeId);
                maxVisitorCount += (facilityData as FacilityData).Capacity;
                EntertainmentData bData = (EntertainmentData)BuildingData.GetBuilding(facility.BuildingTypeId);
                if (maxAdrenalin < bData.Adrenalin) maxAdrenalin = bData.Adrenalin;
                if (minAdrenalin > bData.Adrenalin) minAdrenalin = bData.Adrenalin;
            }
            maxVisitorCount = maxVisitorCount <= 100 ? maxVisitorCount : 100;

            if (_entertainmentFacilities.Count != 0 && GameMap.Time % 2 == 0)
            {
                Int32 chance = 150;
                if (maxVisitorCount > GameMap.Visitors.Count * 10) chance *= 2;
                if (maxVisitorCount <= GameMap.Visitors.Count) chance /= 10;
                if (maxVisitorCount * 0.8 < GameMap.Visitors.Count) chance /= 2;
                chance += _entertainmentFacilities.Count * 15;
                if ((_entertainmentFacilities.Count + _cateringFacilities.Count) * 25 < GameMap.Ticket) chance /= 4;
                if (minAdrenalin == maxAdrenalin && _entertainmentFacilities.Count >= 2) chance /= 4;
                if (maxVisitorCount * 1.4 < GameMap.Visitors.Count) chance = 0;

                Random random = new Random();
                if (random.Next(0,1000) <= chance)
                {
                    Visitor ghost = new Visitor(++_personId, new Tuple<int, int>(GameMap.Buildings[0].Coord.Item1 + 1, GameMap.Buildings[0].Coord.Item2 + 2),
                        new Tuple<int, int>(25, 25));
                    ghost.CurrentActivity = Activity.ARRIVED_TO_ROAD;
                    ghost.ChooseStatus();
                    ghost.Path = PathToDest(ghost.Coord, ChooseDest(ghost));
                    GameMap.Visitors.Add(ghost);
                    GameMap.Money += GameMap.Ticket;
                    OnPersonAdded(ghost);
                }
            }
        }

        /// <summary>
        /// Choose a destination for a person
        /// </summary>
        /// <param name="person">Selected person</param>
        /// <returns>Unique building id</returns>
        private Int32 ChooseDest(Person person)
        {
            if(person is Visitor)
            {
                Visitor visitor = person as Visitor;
                Random random = new Random();
                switch (visitor.CurrentStatus)
                {
                    case Status.HUNGRY:

                        if (visitor.CurrentActivity == Activity.WANDERING)
                        {
                            if (_roadIds.Count > 0) return (_roadIds[random.Next(0, _roadIds.Count)]);
                            else return 0;
                        }

                        var CateringFacilities100 = _cateringFacilities.Where(x => (x.Fee <= 100 && x.Fee > 0)).ToList();
                        var CateringFacilities250 = _cateringFacilities.Where(x => (x.Fee <= 250 && x.Fee > 100)).ToList();
                        var CateringFacilities500 = _cateringFacilities.Where(x => (x.Fee <= 500 && x.Fee > 250)).ToList();
                        var CateringFacilities750 = _cateringFacilities.Where(x => (x.Fee <= 750 && x.Fee > 500)).ToList();
                        var CateringFacilities1000 = _cateringFacilities.Where(x => (x.Fee <= 1000 && x.Fee > 750)).ToList();

                        if (visitor.Money <= 100)
                        {
                            if (_cateringFacilities.Count == 0) return 0;
                            else if (CateringFacilities100.Count > 0) return CateringFacilities100[random.Next(0, CateringFacilities100.Count)].BuildingId;
                            else
                            {
                                visitor.CurrentStatus = Status.NOT_ENOUGH_MONEY;
                                visitor.CurrentActivity = Activity.LEAVING;
                                return -2;
                            }
                        }
                        else if (visitor.Money <= 250)
                        {
                            if (_cateringFacilities.Count == 0) return 0;
                            else if (CateringFacilities250.Count > 0) return CateringFacilities250[random.Next(0, CateringFacilities250.Count)].BuildingId;
                            else if (CateringFacilities100.Count > 0) return CateringFacilities100[random.Next(0, CateringFacilities100.Count)].BuildingId;
                            else
                            {
                                visitor.CurrentStatus = Status.NOT_ENOUGH_MONEY;
                                visitor.CurrentActivity = Activity.LEAVING;

                                return -2;
                            }
                        }
                        else if (visitor.Money <= 500)
                        {
                            if (_cateringFacilities.Count == 0) return 0;
                            else if (CateringFacilities500.Count > 0) return CateringFacilities500[random.Next(0, CateringFacilities500.Count)].BuildingId;
                            else if (CateringFacilities250.Count > 0) return CateringFacilities250[random.Next(0, CateringFacilities250.Count)].BuildingId;
                            else if (CateringFacilities100.Count > 0) return CateringFacilities100[random.Next(0, CateringFacilities100.Count)].BuildingId;
                            else
                            {
                                visitor.CurrentStatus = Status.NOT_ENOUGH_MONEY;
                                visitor.CurrentActivity = Activity.LEAVING;
                                return -2;
                            }
                        }
                        else if (visitor.Money <= 750)
                        {
                            if (_cateringFacilities.Count == 0) return 0;
                            else if (CateringFacilities750.Count > 0) return CateringFacilities750[random.Next(0, CateringFacilities750.Count)].BuildingId;
                            else if (CateringFacilities500.Count > 0) return CateringFacilities500[random.Next(0, CateringFacilities500.Count)].BuildingId;
                            else if (CateringFacilities250.Count > 0) return CateringFacilities250[random.Next(0, CateringFacilities250.Count)].BuildingId;
                            else if (CateringFacilities100.Count > 0) return CateringFacilities100[random.Next(0, CateringFacilities100.Count)].BuildingId;
                            else
                            {
                                visitor.CurrentStatus = Status.NOT_ENOUGH_MONEY;
                                visitor.CurrentActivity = Activity.LEAVING;
                                return -2;
                            }
                        }
                        else
                        {
                            if (_cateringFacilities.Count == 0) return 0;
                            else if (CateringFacilities1000.Count > 0) return CateringFacilities1000[random.Next(0, CateringFacilities1000.Count)].BuildingId;
                            else if (CateringFacilities750.Count > 0) return CateringFacilities750[random.Next(0, CateringFacilities750.Count)].BuildingId;
                            else if (CateringFacilities500.Count > 0) return CateringFacilities500[random.Next(0, CateringFacilities500.Count)].BuildingId;
                            else if (CateringFacilities250.Count > 0) return CateringFacilities250[random.Next(0, CateringFacilities250.Count)].BuildingId;
                            else if (CateringFacilities100.Count > 0) return CateringFacilities100[random.Next(0, CateringFacilities100.Count)].BuildingId;
                            else
                            {
                                visitor.CurrentStatus = Status.NOT_ENOUGH_MONEY;
                                visitor.CurrentActivity = Activity.LEAVING;
                                return -2;
                            }
                        }

                    case Status.UNHAPPY:

                        if (visitor.CurrentActivity == Activity.WANDERING)
                        {
                            if (_roadIds.Count > 0) return (_roadIds[random.Next(0, _roadIds.Count)]);
                            else return 0;
                        }

                        List<Tuple<Int32, List<EntertainmentFacility>>> pairs = new List<Tuple<int, List<EntertainmentFacility>>>();

                        if (_entertainmentFacilities.Count > 0)
                        {
                            foreach (EntertainmentFacility entertainmentFacility in _entertainmentFacilities)
                            {
                                Int32 adrenalin = Math.Abs((BuildingData.GetBuilding(entertainmentFacility.BuildingTypeId) as EntertainmentData).Adrenalin - visitor.AdrenalinLevel);
                                if (pairs.Count > 0)
                                {
                                    Tuple<Int32, List<EntertainmentFacility>> tuple = pairs.SingleOrDefault(x => x.Item1 == adrenalin);
                                    if (tuple != null)
                                    {
                                        tuple.Item2.Add(entertainmentFacility);
                                    }
                                    else
                                    {
                                        List<EntertainmentFacility> list = new List<EntertainmentFacility>();
                                        list.Add(entertainmentFacility);
                                        pairs.Add(new Tuple<Int32, List<EntertainmentFacility>>(adrenalin, list));
                                    }
                                }
                                else
                                {
                                    List<EntertainmentFacility> list = new List<EntertainmentFacility>();
                                    list.Add(entertainmentFacility);
                                    pairs.Add(new Tuple<Int32, List<EntertainmentFacility>>(adrenalin, list));
                                }
                            }

                            List<Tuple<Int32, List<EntertainmentFacility>>> sortedPairs = pairs.OrderBy(x => x.Item1).ToList();

                            foreach (Tuple<Int32, List<EntertainmentFacility>> pair in sortedPairs)
                            {
                                EntertainmentFacility fac = pair.Item2[random.Next(0, pair.Item2.Count)];
                                List<EntertainmentFacility> selected = new List<EntertainmentFacility>();
                                if (visitor.Money - fac.Fee >= 0)
                                {
                                    return fac.BuildingId;
                                }
                                else
                                {
                                    selected.Add(fac);
                                    while (selected.Count != pair.Item2.Count)
                                    {
                                        fac = pair.Item2[random.Next(0, pair.Item2.Count)];
                                        while (selected.Contains(fac))
                                        {
                                            fac = pair.Item2[random.Next(0, pair.Item2.Count)];
                                        }
                                        if (visitor.Money - fac.Fee >= 0)
                                        {
                                            return fac.BuildingId;
                                        }
                                        else
                                        {
                                            selected.Add(fac);
                                        }
                                    }
                                }
                            }

                            visitor.CurrentStatus = Status.NOT_ENOUGH_MONEY;
                            visitor.CurrentActivity = Activity.LEAVING;
                            return -2;

                        }
                        else
                        {
                            return 0;
                        }


                    case Status.SATISFIED:

                        if (_roadIds.Count > 0) return (_roadIds[random.Next(0, _roadIds.Count)]);
                        else return 0;

                    case Status.NOT_ENOUGH_MONEY:

                        return -2;

                    case Status.SAD:

                        return -2;
                }

                return 0;
            }
            else
            {
                Repairman repairman = person as Repairman;
                Random random = new Random();
                if (_roadIds.Count > 0) return (_roadIds[random.Next(0, _roadIds.Count)]);
                else return 0;
            }

        }

        /// <summary>
        /// Sends visitors back home
        /// </summary>
        private void SendHomeVisitors()
        {
            if(GameMap.Visitors != null)
            {
                Stack<Tuple<int, int>> path = null;
                foreach (Visitor visitor in GameMap.Visitors)
                {
                    if(visitor.Coord.Item1 == 14 && visitor.Coord.Item2 == 7)
                    {
                        visitor.CurrentActivity = Activity.ARRIVED_TO_EXIT;
                    }
                    else
                    {
                        visitor.CurrentActivity = Activity.LEAVING;
                        path = PathToDest(visitor.Coord, -2);
                        visitor.Path = path;
                    }
                }
                GameSpeed = Speed.FAST;
            }
        }

        /// <summary>
        /// Counts the path to the destination
        /// </summary>
        /// <param name="current">Current coordinates of a person</param>
        /// <param name="destId">Destination</param>
        /// <returns>Path to the destination</returns>
        private Stack<Tuple<Int32, Int32>> PathToDest(Tuple<Int32, Int32> current, Int32 destId)
        {
            if (destId == 0) return null;

            Building start = _map[current.Item1, current.Item2];
            if (start.BuildingId == destId) return null;

            var parents = new Dictionary<Building, Building>();
            var distances = new Dictionary<Building, Int32>();
            var queue = new Queue<Building>();

            foreach(Building building in GameMap.Buildings)
            {
                if(building.BuildingTypeId == -1)
                {
                    Building fake1 = _map[building.Coord.Item1, building.Coord.Item2 + 2];
                    Building fake2 = _map[building.Coord.Item1 + 1, building.Coord.Item2 + 2];
                    parents[fake1] = null;
                    distances[fake1] = -1;
                    parents[fake2] = null;
                    distances[fake2] = -1;
                } else if(start.BuildingTypeId > 0 && building.BuildingId == start.BuildingId)
                {
                    Building fake = new Road(-3, 0, current);
                    start = fake;
                    parents[fake] = null;
                    distances[fake] = -1;
                } else
                {
                    parents[building] = null;
                    distances[building] = -1;
                }
            }

            distances[start] = 0;
            queue.Enqueue(start);
            
            Int32 y, x;
            while(queue.Count != 0)
            {
                Building b = queue.Dequeue();
                
                y = b.Coord.Item1;
                x = b.Coord.Item2;

                if (CheckNeighbour(y + 1, x, b, parents, distances, queue, destId)) return MakePath(_map[y + 1,x], y + 1, x, parents);
                if (CheckNeighbour(y - 1, x, b, parents, distances, queue, destId)) return MakePath(_map[y - 1, x], y - 1, x, parents);
                if (CheckNeighbour(y, x + 1, b, parents, distances, queue, destId)) return MakePath(_map[y, x + 1], y, x + 1, parents);
                if (CheckNeighbour(y, x - 1, b, parents, distances, queue, destId)) return MakePath(_map[y, x - 1], y, x - 1,parents);
            }

            return null;
        }

        /// <summary>
        /// Makes the path to destination
        /// </summary>
        /// <param name="dest">Destination</param>
        /// <param name="y">Horiontal coordinate</param>
        /// <param name="x"> Vertical coordinate</param>
        /// <param name="parents">So far selected path</param>
        /// <returns>So far selected path to destination</returns>
        private Stack<Tuple<Int32, Int32>> MakePath(Building dest, Int32 y, Int32 x, Dictionary<Building, Building> parents)
        {
            var stack = new Stack<Tuple<Int32, Int32>>();
            Building b = dest;

            if(b.BuildingTypeId != 0)
            {
                stack.Push(new Tuple<int, int>(y, x));
                b = parents[b];
            }

            while (parents[b] != null)
            {
                stack.Push(b.Coord);
                b = parents[b];
            }

            return stack;
        }

        /// <summary>
        /// Checks the neighbour of a field
        /// </summary>
        /// <param name="y">Vertical index</param>
        /// <param name="x">Horizontal index</param>
        /// <param name="parent">Parent field</param>
        /// <param name="parents">So far selected path</param>
        /// <param name="distances">Distances</param>
        /// <param name="queue">Buildings in queue</param>
        /// <param name="id">Destination</param>
        /// <returns>Whether the path can be made </returns>
        private Boolean CheckNeighbour(Int32 y, Int32 x, Building parent, Dictionary<Building, Building> parents,
                                        Dictionary<Building, Int32> distances, Queue<Building> queue, Int32 id)
        {
            if (y >= MapSize || y < 0 || x >= MapSize || x < 0) return false;

            Building b = _map[y, x];
            if (b is null || b.BuildingId == 1) return false;
            if (b.BuildingId != id && !(b is Road)) return false;
            if (distances[b] != -1) return false;

            distances[b] = distances[parent] + 1;
            parents[b] = parent;
            queue.Enqueue(b);

            if (b.BuildingId == id) return true;

            return false;
        }

        /// <summary>
        /// Joins a visitor to an entertainment facility
        /// </summary>
        /// <param name="visitor">Visitor</param>
        /// <param name="ent">Entertainment facility</param>
        public void JoinEntertainment(Visitor visitor, EntertainmentFacility ent)
        {
            var entStatic = BuildingData.GetBuilding(ent.BuildingTypeId) as EntertainmentData;
            if (ent.Status == FacilityStatus.OPERATING)
            {
                ent.VisitorQueue.Enqueue(visitor);
                visitor.CurrentActivity = Activity.WAITING;
            }
            else if (ent.UserList.Count < entStatic.Capacity)
            {
                ent.UserList.Enqueue(visitor);
                visitor.CurrentActivity = Activity.WAITING;
                ent.Status = FacilityStatus.WAITING;
            }
            else
            {
                ent.VisitorQueue.Enqueue(visitor);
                visitor.CurrentActivity = Activity.WAITING;
                ent.Status = FacilityStatus.WAITING;
            }
        }

        /// <summary>
        /// Refreshes entertainment facility
        /// </summary>
        /// <param name="ent">Entertainment facility</param>
        public void RefreshEntertainment(EntertainmentFacility ent)
        {
            var entStatic = BuildingData.GetBuilding(ent.BuildingTypeId) as EntertainmentData;
            if (ent.Status == FacilityStatus.WAITING)
            {
                if (ent.UserList.Count >= ent.MinimumCapacity)
                {
                    foreach (Visitor visitor in ent.UserList)
                    {
                        visitor.CurrentActivity = Activity.USING;
                        visitor.CurrentInteractionTimeLeft = entStatic.InteractionLength;
                    }
                    ent.Status = FacilityStatus.OPERATING;
                }
            }
            else if (ent.Status == FacilityStatus.OPERATING)
            {
                bool ended = false;
                foreach (Visitor visitor in ent.UserList)
                {
                    if (visitor.CurrentInteractionTimeLeft == 0)
                    {
                        visitor.Money -= ent.Fee;
                        visitor.HappinessLevel += entStatic.Happiness;
                        ended = true;
                        visitor.ChooseStatus();
                    }
                }
                if (ended)
                {
                    int n = ent.UserList.Count;
                    for (int i = 0; i < n; ++i)
                    {

                        Visitor visitor = ent.UserList.Dequeue();
                    }

                    while (ent.UserList.Count < entStatic.Capacity && ent.VisitorQueue.Count > 0)
                    {
                        Visitor visitor = ent.VisitorQueue.Dequeue();
                        ent.UserList.Enqueue(visitor);
                    }

                    ++ent.RoundCount;
                    if (ent.RoundCount % 5 == 0)
                    {
                        ent.Status = FacilityStatus.FAULTY;
                    }
                    else
                    {
                        ent.Status = FacilityStatus.WAITING;
                    }
                }
            }
        }

        /// <summary>
        /// Joins a visitor to a catering facility
        /// </summary>
        /// <param name="visitor">Visitor</param>
        /// <param name="cat">Catering facility</param>
        public void JoinCatering(Visitor visitor, CateringFacility cat)
        {
            var catStatic = BuildingData.GetBuilding(cat.BuildingTypeId) as CateringData;
            if (cat.UserList.Count < catStatic.Capacity && cat.VisitorQueue.Count == 0)
            {
                cat.UserList.Enqueue(visitor);
                visitor.CurrentActivity = Activity.USING;
                visitor.CurrentInteractionTimeLeft = catStatic.InteractionLength;
                cat.Status = FacilityStatus.OPERATING;
            }
            else
            {
                cat.VisitorQueue.Enqueue(visitor);
                visitor.CurrentActivity = Activity.WAITING;
            }
        }

        /// <summary>
        /// Refreshes a catering facility
        /// </summary>
        /// <param name="cat">Catering facility</param>
        public void RefreshCatering(CateringFacility cat)
        {
            var catStatic = BuildingData.GetBuilding(cat.BuildingTypeId) as CateringData;
            int i = 0;
            foreach (Visitor visitor in cat.UserList)
            {
                if (visitor.CurrentInteractionTimeLeft == 0)
                {
                    visitor.Money -= cat.Fee;
                    visitor.FoodLevel += catStatic.Food;
                    visitor.ChooseStatus();
                    ++i;
                }
            }

            for (int j = 0; j < i; ++j)
            {
                cat.UserList.Dequeue();
            }

            if (cat.VisitorQueue.Count > 0)
            {
                while (cat.UserList.Count < catStatic.Capacity && cat.VisitorQueue.Count > 0)
                {
                    Visitor visitor = cat.VisitorQueue.Dequeue();
                    visitor.CurrentActivity = Activity.USING;
                    visitor.CurrentInteractionTimeLeft = catStatic.InteractionLength;
                    cat.UserList.Enqueue(visitor);
                }
            }
            cat.Status = (cat.UserList.Count == 0) ? FacilityStatus.WAITING : FacilityStatus.OPERATING;
        }

        #endregion

        #region Persistence methods

        /// <summary>
        /// Loads a game
        /// </summary>
        /// <param name="path">File path</param>
        public async Task LoadGameAsync(String path)
        {
            _roadIds = new List<int>();
            _entertainmentIds = new List<int>();
            _cateringIds = new List<int>();
            _entertainmentFacilities = new List<EntertainmentFacility>();
            _cateringFacilities = new List<CateringFacility>();
            _map = new Building[MapSize, MapSize];
            for (int i = 0; i < MapSize; ++i)
            {
                for (int j = 0; j < MapSize; ++j)
                {
                    _map[i, j] = null;
                }
            }
            _personId = _buildingId = 1;
            if (_dataAccess == null)
            {
                throw new InvalidOperationException("No data access is provided");
            }
            GameMap = await _dataAccess.LoadAsync(path);
            _gameSpeed = Speed.PAUSE;
            foreach (Building building in GameMap.Buildings)
            {
                StaticBuildingData bData = BuildingData.GetBuilding(building.BuildingTypeId);

                for (int i = building.Coord.Item1; i < building.Coord.Item1 + bData.Size.Item1; i++)
                {
                    for (int j = building.Coord.Item2; j < building.Coord.Item2 + bData.Size.Item2; j++)
                    {
                        _map[i, j] = building;
                    }
                }
                if (building is Gate)
                {
                    _map[building.Coord.Item1, building.Coord.Item2 + 2] = new Road(-1, 0, new Tuple<int, int>(building.Coord.Item1, building.Coord.Item2 + 2));
                    _map[building.Coord.Item1 + 1, building.Coord.Item2 + 2] = new Road(-2, 0, new Tuple<int, int>(building.Coord.Item1 + 1, building.Coord.Item2 + 2));
                }

                if (building.BuildingId > _buildingId) _buildingId = building.BuildingId;

                if (building is Road) _roadIds.Add(building.BuildingId);
                else if (building is EntertainmentFacility)
                {
                    _entertainmentIds.Add(building.BuildingId);
                    _entertainmentFacilities.Add(building as EntertainmentFacility);
                }
                else if (building is CateringFacility)
                {
                    _cateringIds.Add(building.BuildingId);
                    _cateringFacilities.Add(building as CateringFacility);
                }
            }
            _buildingId++;
            _personId++;
            foreach (Building building in GameMap.Buildings)
            {
                OnBuildingAdded(building);
                if (building is Facility)
                {
                    OnBuildingChanged(building);
                }
            }
            ParkStatus = ParkStatus.CLOSED;
            foreach (Person person in GameMap.Visitors)
            {
                if (person.PersonId > _personId) _personId = person.PersonId;
            }
            foreach (Person person in GameMap.Repairmen)
            {
                if (person.PersonId > _personId) _personId = person.PersonId;
            }
            
            OnNewGameStarted(GameMap, GameMap.Name);
            OnTimePassed();
        }

        /// <summary>
        /// Saves a game
        /// </summary>
        /// <param name="path">File path</param>
        public async Task SaveGameAsync(String path)
        {
            if (_dataAccess == null)
            {
                throw new InvalidOperationException("No data access is provided");
            }
            await _dataAccess.SaveAsync(path, GameMap);
        }

        #endregion

        #region Private event methods

        /// <summary>
        /// Event method for canvas refresh
        /// </summary>
        private void OnVisitorsChanged()
        {
            VisitorsChanged?.Invoke(this, EventArgs.Empty);
        }

        /// <summary>
        /// Event method for person change
        /// </summary>
        /// <param name="person">Person</param>
        private void OnPersonChanged(Person person)
        {
            PersonChanged?.Invoke(this, new PersonChangedEventArgs(person));
        }

        /// <summary>
        /// Event method for person add to GameMap
        /// </summary>
        /// <param name="person"></param>
        private void OnPersonAdded(Person person)
        {
            PersonAdded?.Invoke(this, new PersonChangedEventArgs(person));
        }

        /// <summary>
        /// Event methos for person delete from GameMap
        /// </summary>
        /// <param name="person">Person</param>
        private void OnPersonDeleted(Person person)
        {
            PersonDeleted?.Invoke(this, new PersonChangedEventArgs(person));
        }

        /// <summary>
        /// Event method for building add to GameMap
        /// </summary>
        /// <param name="building">Building</param>
        private void OnBuildingAdded(Building building)
        {
            BuildingAdded?.Invoke(this, new BuildingChangedEventArgs(building));
        }

        /// <summary>
        /// Event method for building delete to GameMap
        /// </summary>
        /// <param name="building">Building</param>
        private void OnBuildingDestroyed(Building building)
        {
            BuildingDestroyed?.Invoke(this, new BuildingChangedEventArgs(building));
        }

        /// <summary>
        /// Event method for building change
        /// </summary>
        /// <param name="building">Building</param>
        private void OnBuildingChanged(Building building)
        {
            BuildingChanged?.Invoke(this, new BuildingChangedEventArgs(building));
        }

        /// <summary>
        /// Event method for new game start
        /// </summary>
        /// <param name="gameMap">GameMap</param>
        /// <param name="name">Park name</param>
        private void OnNewGameStarted(GameMap gameMap, String name)
        {
            NewGameStarted?.Invoke(this, new NewGameEventArgs(gameMap, name));
        }

        /// <summary>
        /// Event method for speed change
        /// </summary>
        /// <param name="gameSpeed">Game speed</param>
        private void OnSpeedChanged(Speed gameSpeed)
        {
            SpeedChanged?.Invoke(this, new SpeedChangedEventArgs(gameSpeed));
        }

        /// <summary>
        /// Event method for tick
        /// </summary>
        private void OnTimePassed()
        {
            TimePassed?.Invoke(this, EventArgs.Empty);
        }

        /// <summary>
        /// Event method for money change
        /// </summary>
        private void OnMoneyChanged()
        {
            MoneyChanged?.Invoke(this, EventArgs.Empty);
        }

        /// <summary>
        /// Event method for park status change
        /// </summary>
        private void OnParkStatusChanged()
        {
            ParkStatusChanged?.Invoke(this, EventArgs.Empty);
        }

        /// <summary>
        /// Event method for park close
        /// </summary>
        private void OnEnableDestroy()
        {
            EnableDestroy?.Invoke(this, EventArgs.Empty);
        }

        /// <summary>
        /// Event method for info refresh
        /// </summary>
        private void OnPopupRefreshed()
        {
            PopupRefreshed?.Invoke(this, EventArgs.Empty);
        }

        #endregion
    }
}
