using Microsoft.VisualStudio.TestTools.UnitTesting;
using HELLo_World.Model;
using HELLo_World.MapData;
using HELLo_World.Persistence;
using Moq;
using HELLo_World.MapData.Buildings;
using System;
using System.Threading.Tasks;
using HELLo_World.Model.ModelExceptions;
using HELLo_World.Model.Information;
using HELLo_World.Model.EventArguments;

namespace HELLo_World.Test
{
    [TestClass]
    public class GameTest
    {
        #region Fields

        private GameModel _model;
        private GameMap _mockedMap;
        private Mock<IGamePersistence> _mock;

        #endregion

        #region Initialization

        [TestInitialize]
        public void Initialize()
        {
            _mockedMap = new GameMap("test");
            _mockedMap.Buildings.Add(new Gate(1, -1, new Tuple<int, int>(13,5)));
            _mockedMap.Buildings.Add(new Road(2, 0, new Tuple<int, int>(12, 7)));
            _mockedMap.Buildings.Add(new Road(3, 0, new Tuple<int, int>(11, 7)));
            _mockedMap.Buildings.Add(new Road(4, 0, new Tuple<int, int>(10, 7)));
            _mockedMap.Buildings.Add(new Road(5, 0, new Tuple<int, int>(9, 7)));
            _mockedMap.Buildings.Add(new CateringFacility(6, 2, new Tuple<int, int>(9, 5), 30));
            _mockedMap.Buildings.Add(new EntertainmentFacility(7, 102, new Tuple<int, int>(9, 8), 20, 5));
            _mockedMap.Money = 4500;
            _mockedMap.Ticket = 75;
            _mockedMap.Time = 100;

            _mock = new Mock<IGamePersistence>();
            _mock.Setup(mock => mock.LoadAsync(It.IsAny<String>())).Returns(() => Task.FromResult(_mockedMap));

            _model = new GameModel(_mock.Object);
            _model.NewGameStarted += new EventHandler<NewGameEventArgs>(Model_NewGameStarted);
            _model.SpeedChanged += new EventHandler<SpeedChangedEventArgs>(Model_SpeedChanged);
        }

        #endregion

        #region NewGame tests

        [TestMethod]
        public void NewGameTest()
        {
            _model.NewGame("test");

            Assert.AreEqual(0, _model.NumberOfRepairmen);
            Assert.AreEqual(1, _model.GameMap.Buildings.Count);
            Assert.AreEqual(Speed.PAUSE, _model.GameSpeed);
            Assert.AreEqual(0, _model.GameMap.Visitors.Count);
            Assert.AreEqual("test", _model.GameMap.Name);
            Assert.AreEqual(ParkStatus.CLOSED, _model.ParkStatus);
        }

        #endregion

        #region Speed tests

        [TestMethod]
        public void GameSpeedTestStart()
        {
            _model.NewGame("test");
            Assert.AreEqual(Speed.PAUSE, _model.GameSpeed);
            _model.GameSpeed = Speed.NORMAL;

            Assert.AreEqual(Speed.NORMAL, _model.GameSpeed);
        }

        [TestMethod]
        public void GameSpeedTestPause()
        {
            _model.NewGame("test");
            _model.GameSpeed = Speed.NORMAL;
            _model.GameSpeed = Speed.PAUSE;
            Assert.AreEqual(Speed.PAUSE, _model.GameSpeed);
        }

        [TestMethod]
        public void GameSpeedTestSpeedUp()
        {
            _model.NewGame("test");
            _model.GameSpeed = Speed.NORMAL;
            _model.GameSpeed = Speed.FAST;
            Assert.AreEqual(Speed.FAST, _model.GameSpeed);
        }

        [TestMethod]
        public void GameSpeedTestSlowDown()
        {
            _model.NewGame("test");
            _model.GameSpeed = Speed.NORMAL;
            _model.GameSpeed = Speed.SLOW;
            Assert.AreEqual(Speed.SLOW, _model.GameSpeed);
        }

        #endregion

        #region Repairmen tests

        [TestMethod]
        public void NumberOfRepairmenIncrease()
        {
            _model.NewGame("test");

            _model.NumberOfRepairmen = 3;
            Assert.AreEqual(3, _model.NumberOfRepairmen);
            foreach (var repairmen in _model.GameMap.Repairmen)
            {
                Assert.AreEqual(repairmen.Coord.Item1, 14);
                Assert.AreEqual(repairmen.Coord.Item2, 7);
            }
            _model.NumberOfRepairmen = 5;
            Assert.AreEqual(5, _model.NumberOfRepairmen);
        }

        [TestMethod]
        public void NumberOfRepairmenDecrease()
        {
            _model.NewGame("test");

            _model.NumberOfRepairmen = 3;
            Assert.AreEqual(3, _model.NumberOfRepairmen);
            _model.NumberOfRepairmen = 1;
            Assert.AreEqual(1, _model.NumberOfRepairmen);
        }

        [TestMethod]
        public void NumberOfRepairmenSetToZero()
        {
            _model.NewGame("test");

            _model.NumberOfRepairmen = 3;
            Assert.AreEqual(3, _model.NumberOfRepairmen);
            _model.NumberOfRepairmen = 0;
            Assert.AreEqual(0, _model.NumberOfRepairmen);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentException))]
        public void NumberOfRepairmenInvalidNumber()
        {
            _model.NewGame("test");
            _model.NumberOfRepairmen = -3;
        }

        #endregion

        #region ParkStatus tests

        [TestMethod]
        public void ParkStatusSetOpen()
        {
            _model.NewGame("test");
            _model.ParkStatus = ParkStatus.OPEN;

            Assert.AreEqual(ParkStatus.OPEN, _model.ParkStatus);
        }

        [TestMethod]
        public void ParkStatusSetClosed()
        {
            _model.NewGame("test");
            _model.ParkStatus = ParkStatus.OPEN;
            _model.ParkStatus = ParkStatus.CLOSED;

            Assert.AreEqual(ParkStatus.CLOSED, _model.ParkStatus);
        }

        [TestMethod]
        public void ParkStatusSetClosedVisitorsLeaving()
        {
            _model.NewGame("test");
            _model.ParkStatus = ParkStatus.OPEN;
            _model.GameMap.Visitors.Add(new MapData.People.Visitor(1, new Tuple<int, int>(14, 7), new Tuple<int, int>(25, 25)));
            _model.GameMap.Visitors.Add(new MapData.People.Visitor(2, new Tuple<int, int>(13, 7), new Tuple<int, int>(12, 45)));
            _model.ParkStatus = ParkStatus.CLOSED;

            Assert.AreEqual(MapData.People.Activity.ARRIVED_TO_EXIT, _model.GameMap.Visitors[0].CurrentActivity);
            Assert.AreEqual(MapData.People.Activity.LEAVING, _model.GameMap.Visitors[1].CurrentActivity);
        }

        #endregion

        #region Build tests

        [TestMethod]
        public void BuildRoadTest()
        {
            _model.NewGame("test");
            int prevMoney = _model.GameMap.Money;

            _model.Build(new Tuple<int, int>(1,1), 0);

            Assert.AreEqual(2, _model.GameMap.Buildings.Count);
            Assert.AreEqual(0, _model[1,1]);
            Assert.AreEqual(prevMoney - 50, _model.GameMap.Money);
        }

        [TestMethod]
        public void BuildBiggerTest()
        {
            _model.NewGame("test");
            int prevMoney = _model.GameMap.Money;

            _model.Build(new Tuple<int, int>(1, 1), 1);

            Assert.AreEqual(2, _model.GameMap.Buildings.Count);
            Assert.AreEqual(1, _model[1, 1]);
            Assert.AreEqual(1, _model[1, 2]);
            Assert.AreEqual(1, _model[2, 1]);
            Assert.AreEqual(1, _model[2, 2]);
            Assert.AreEqual(prevMoney - 500, _model.GameMap.Money);
        }

        [TestMethod]
        public void BuildNotEnoughMoneyTest()
        {
            _model.NewGame("test");
            _model.GameMap.Money = 0;
            try
            {
                _model.Build(new Tuple<int, int>(0,0), 0);
                Assert.Fail(); // If it gets to this line, no exception was thrown
            }
            catch (BuildException e) 
            {
                Assert.AreEqual(BuildErrorType.NOT_ENOUGH_MONEY, e.ErrorType);
            }
        }

        [TestMethod]
        public void BuildOffMapTest()
        {
            _model.NewGame("test");
            try
            {
                _model.Build(new Tuple<int, int>(GameModel.MapSize - 1, GameModel.MapSize - 1), 1);
                Assert.Fail(); // If it gets to this line, no exception was thrown
            }
            catch (BuildException e)
            {
                Assert.AreEqual(BuildErrorType.OFF_MAP, e.ErrorType);
            }
        }

        [TestMethod]
        public void BuildOccupiedTest()
        {
            _model.NewGame("test");
            _model.Build(new Tuple<int, int>(1, 1), 0);
            try
            {
                _model.Build(new Tuple<int, int>(0, 0), 1);
                Assert.Fail(); // If it gets to this line, no exception was thrown
            }
            catch (BuildException e)
            {
                Assert.AreEqual(BuildErrorType.SPACE_OCCUPIED, e.ErrorType);
            }
        }

        #endregion

        #region Destroy tests

        [TestMethod]
        public void DestroyRoadTest()
        {
            _model.NewGame("test");
            _model.Build(new Tuple<int, int>(1,1), 0);
            _model.Destroy(new Tuple<int, int>(1,1), new Tuple<int, int>(1,1));

            Assert.AreEqual(1, _model.GameMap.Buildings.Count);
            Assert.AreEqual(-1, _model[1,1]);
        }

        [TestMethod]
        public void DestroyBiggerTest()
        {
            _model.NewGame("test");
            _model.Build(new Tuple<int, int>(1, 1), 1);
            _model.Destroy(new Tuple<int, int>(1, 1), new Tuple<int, int>(2, 2));

            Assert.AreEqual(1, _model.GameMap.Buildings.Count);
            Assert.AreEqual(-1, _model[1, 1]);
            Assert.AreEqual(-1, _model[1, 2]);
            Assert.AreEqual(-1, _model[2, 1]);
            Assert.AreEqual(-1, _model[2, 2]);
        }

        [TestMethod]
        public void DestroyNoBuildingTest()
        {
            _model.NewGame("test");
            try
            {
                _model.Destroy(new Tuple<int, int>(0, 0), new Tuple<int, int>(1,1));
                Assert.Fail(); // If it gets to this line, no exception was thrown
            }
            catch (DestroyException e)
            {
                Assert.AreEqual(DestroyErrorType.NO_BUILDING, e.ErrorType);
            }
        }

        [TestMethod]
        public void DestroyGateTest()
        {
            _model.NewGame("test");
            try
            {
                _model.Destroy(new Tuple<int, int>(13, 5), new Tuple<int, int>(2, 5));
                Assert.Fail(); // If it gets to this line, no exception was thrown
            }
            catch (DestroyException e)
            {
                Assert.AreEqual(DestroyErrorType.GATE, e.ErrorType);
            }
        }

        [TestMethod]
        public void DestroyWhenNotEmptyTest()
        {
            _model.NewGame("test");
            _model.Build(new Tuple<int, int>(0,0), 0);
            _model.GameMap.Visitors.Add(new MapData.People.Visitor(1, new Tuple<int, int>(0,0), new Tuple<int, int>(25,25)));
            try
            {
                _model.Destroy(new Tuple<int, int>(0, 0), new Tuple<int, int>(1, 1));
                Assert.Fail(); // If it gets to this line, no exception was thrown
            }
            catch (DestroyException e)
            {
                Assert.AreEqual(DestroyErrorType.NOT_EMPTY, e.ErrorType);
            }
        }

        #endregion

        #region GetInfo tests

        [TestMethod]
        public void GetBuidlingInfoFacilityTest()
        {
            _model.NewGame("test");
            _model.Build(new Tuple<int, int>(1, 1), 1);
            BuildingInfo info = _model.GetBuildingInfo(2);

            Assert.AreEqual(2, info.BuildingId);
            Assert.AreEqual(5, info.Capacity.Item2);
            Assert.AreEqual(10, info.Fee);
            Assert.AreEqual("HellDog", info.Name);
            Assert.AreEqual(FacilityStatus.UNDER_CONSTRUCTION, info.Status);
        }

        [TestMethod]
        public void GetBuildingInfoRoadTest()
        {
            _model.NewGame("test");
            _model.Build(new Tuple<int, int>(1, 1), 0);
            BuildingInfo info = _model.GetBuildingInfo(2);
            Assert.IsNull(info);
        }

        [TestMethod]
        public void GetBuildingInfoInvalidIdTest()
        {
            _model.NewGame("test");
            BuildingInfo info = _model.GetBuildingInfo(2);
            Assert.IsNull(info);
        }

        [TestMethod]
        public void GetVisitorInfoValidIdTest()
        {
            _model.NewGame("test");
            _model.GameMap.Visitors.Add(new MapData.People.Visitor(3, new Tuple<int, int>(4,5), new Tuple<int, int>(22,33)));
            VisitorInfo info = _model.GetVisitorInfo(3);
            Assert.AreEqual(_model.GameMap.Visitors[0].AdrenalinLevel, info.Adrenalin);
            Assert.AreEqual(_model.GameMap.Visitors[0].FoodLevel, info.Food);
            Assert.AreEqual(_model.GameMap.Visitors[0].CurrentActivity, info.CurrentActivity);
            Assert.AreEqual(_model.GameMap.Visitors[0].CurrentStatus, info.CurrentStatus);
            Assert.AreEqual(_model.GameMap.Visitors[0].HappinessLevel, info.Happiness);
            Assert.AreEqual(_model.GameMap.Visitors[0].Money, info.Money);
        }

        [TestMethod]
        public void GetVisitorInfoInvalIdTest()
        {
            _model.NewGame("test");
            VisitorInfo info = _model.GetVisitorInfo(3);
            Assert.IsNull(info);
        }

        #endregion

        #region AdvanceTime tests

        [TestMethod]
        public void AdvanceTimeTimeChanged()
        {
            _model.NewGame("test");
            _model.Build(new Tuple<int, int>(1,1), 1);
            _model.AdvanceTime();

            Assert.AreEqual(1, _model.GameMap.Time);
            Assert.AreEqual(1, (_model.GameMap.Buildings[1] as Facility).TimeSpentInStatus);
        }

        [TestMethod]
        public void AdvanceTimeVisitorStatusChanged()
        {
            _model.NewGame("test");
            _model.Build(new Tuple<int, int>(12, 7), 0);
            _model.Build(new Tuple<int, int>(11, 7), 0);
            _model.ParkStatus = ParkStatus.OPEN;
            _model.GameSpeed = Speed.FAST;
            _model.GameMap.Visitors.Add(new MapData.People.Visitor(1, new Tuple<int, int>(14,7), new Tuple<int, int>(25,25)));
            _model.AdvanceTime();

            Assert.AreEqual(MapData.People.Activity.WANDERING, _model.GameMap.Visitors[0].CurrentActivity);
        }

        [TestMethod]
        public void AdvanceTimeRepairmenWandering()
        {
            _model.NewGame("test");
            _model.Build(new Tuple<int, int>(12, 7), 0);
            _model.Build(new Tuple<int, int>(11, 7), 0);
            _model.ParkStatus = ParkStatus.OPEN;
            _model.GameSpeed = Speed.FAST;
            _model.NumberOfRepairmen = 1;
            _model.AdvanceTime();
            _model.AdvanceTime();

            Assert.AreEqual(MapData.People.Activity.WANDERING, _model.GameMap.Repairmen[0].CurrentActivity);
            Assert.AreNotEqual(25, _model.GameMap.Repairmen[0].InTileCoord.Item1);
        }

        #endregion

        #region SetTicket method tests

        [TestMethod]
        [ExpectedException(typeof(ArgumentException))]
        public void SetTicketInvalidValueTest()
        {
            _model.NewGame("test");
            _model.SetTicket(-10);
        }

        [TestMethod]
        public void SetTicketValidValueIncrease()
        {
            _model.NewGame("test");
            _model.SetTicket(70);

            Assert.AreEqual(70, _model.GameMap.Ticket);
        }

        [TestMethod]
        public void SetTicketValidValueDecrease()
        {
            _model.NewGame("test");
            _model.SetTicket(20);

            Assert.AreEqual(20, _model.GameMap.Ticket);
        }

        #endregion

        #region SetFee method tests

        [TestMethod]
        public void SetFeeIncreaseFee()
        {
            _model.NewGame("test");
            _model.Build(new Tuple<int, int>(1, 1), 1);

            _model.SetBuildingFee(2, 1111);
            Assert.AreEqual(1111, (_model.GameMap.Buildings[1] as Facility).Fee);
        }

        [TestMethod]
        public void SetFeeDecreaseFee()
        {
            _model.NewGame("test");
            _model.Build(new Tuple<int, int>(1, 1), 1);

            _model.SetBuildingFee(2, 1);
            Assert.AreEqual(1, (_model.GameMap.Buildings[1] as Facility).Fee);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentException))]
        public void SetFeeInvalidIdTest()
        {
            _model.NewGame("test");
            _model.SetBuildingFee(1, 1111);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentException))]
        public void SetFeeInvalidBuildingTypeTest()
        {
            _model.NewGame("test");
            _model.Build(new Tuple<int, int>(1, 1), 0);
            _model.SetBuildingFee(1, 1111);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentException))]
        public void SetFeeInvalidNewValueTest()
        {
            _model.NewGame("test");
            _model.Build(new Tuple<int, int>(1, 1), 1);
            _model.SetBuildingFee(1, -1111);
        }

        #endregion

        #region SetMinimumCapacity tests

        [TestMethod]
        public void SetMinimumCapacityIncreaseCapacity()
        {
            _model.NewGame("test");
            _model.Build(new Tuple<int, int>(1, 1), 101);

            _model.SetMinimumCapacity(2, 9);
            Assert.AreEqual(9, (_model.GameMap.Buildings[1] as EntertainmentFacility).MinimumCapacity);
        }

        [TestMethod]
        public void SetMinimumCapacityDecreaseCapacity()
        {
            _model.NewGame("test");
            _model.Build(new Tuple<int, int>(1, 1), 101);

            _model.SetMinimumCapacity(2, 1);
            Assert.AreEqual(1, (_model.GameMap.Buildings[1] as EntertainmentFacility).MinimumCapacity);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentException))]
        public void SetMinimumCapacityInvalidIdTest()
        {
            _model.NewGame("test");
            _model.SetMinimumCapacity(1, 1111);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentException))]
        public void SetMinimumCapacityInvalidBuildingTypeTest()
        {
            _model.NewGame("test");
            _model.Build(new Tuple<int, int>(1, 1), 0);
            _model.SetMinimumCapacity(1, 1111);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentException))]
        public void SetMinimumCapacityInvalidNewValueTest()
        {
            _model.NewGame("test");
            _model.Build(new Tuple<int, int>(1, 1), 1);
            _model.SetMinimumCapacity(1, -1111);
        }

        #endregion

        #region Persistence method tests

        [TestMethod]
        public async Task LoadGameTest()
        {
            _model.NewGame("t");
            await _model.LoadGameAsync(String.Empty);

            Assert.AreEqual(Speed.PAUSE, _model.GameSpeed);
            Assert.AreEqual(_mockedMap, _model.GameMap);
            Assert.AreEqual(ParkStatus.CLOSED, _model.ParkStatus);
            Assert.AreEqual(7, _model.GameMap.Buildings.Count);
            Assert.AreEqual(4500, _model.GameMap.Money);
            Assert.AreEqual(75, _model.GameMap.Ticket);
            Assert.AreEqual(100, _model.GameMap.Time);
            Assert.AreEqual("test", _model.GameMap.Name);
            Assert.AreEqual(4, _model.GameMap.Buildings[3].BuildingId);
            Assert.AreEqual(0, _model.GameMap.Buildings[3].BuildingTypeId);
            Assert.AreEqual(2, _model.GameMap.Buildings[5].BuildingTypeId);
            Assert.AreEqual(30, (_model.GameMap.Buildings[5] as CateringFacility).Fee);
            Assert.AreEqual(102, _model.GameMap.Buildings[6].BuildingTypeId);
            Assert.AreEqual(20, (_model.GameMap.Buildings[6] as EntertainmentFacility).Fee);
            Assert.AreEqual(5, (_model.GameMap.Buildings[6] as EntertainmentFacility).MinimumCapacity);
        }

        #endregion

        #region Event handlers

        private void Model_NewGameStarted(object sender, NewGameEventArgs e)
        {
            Assert.AreEqual(_model.GameMap, e.GameMap);
        }

        private void Model_SpeedChanged(object sender, SpeedChangedEventArgs e)
        {
            Assert.AreEqual(_model.GameSpeed, e.GameSpeed);
        }

        #endregion
    }
}
