﻿using System;
using System.Collections.Generic;
using System.Configuration;
using System.Data;
using System.Linq;
using System.Threading.Tasks;
using System.Windows;
using HELLo_World.ViewModel;
using HELLo_World.Model;
using HELLo_World.Persistence.Text;
using System.Windows.Threading;
using HELLo_World.Model.EventArguments;
using Microsoft.Win32;
using HELLo_World.Persistence;
using HELLo_World.ViewModel.ViewModels;
using HELLo_World.ViewModel.EventArguments;

namespace HELLo_World.View
{
    public partial class App : Application
    {
        #region Fields

        private MainWindow _window;
        private SettingsWindow _settingsWindow;
        private MainViewModel _viewModel;
        private SettingsViewModel _settingsViewModel;
        private GameModel _model;
        private TextFileGamePersistence _dataAccess;

        private DispatcherTimer _timer;

        private Speed _lastSpeed;

        #endregion

        #region Constructor

        public App()
        {
            Startup += new StartupEventHandler(App_Startup);
        }

        #endregion

        #region App event handler

        private void App_Startup(object sender, StartupEventArgs eventArgs)
        {
            _timer = new DispatcherTimer();
            _timer.Interval = TimeSpan.FromSeconds(0.1);
            _timer.Tick += new EventHandler(Timer_Tick);

            _dataAccess = new TextFileGamePersistence();

            _window = new MainWindow();
            _model = new GameModel(_dataAccess);
            _model.SpeedChanged += new EventHandler<SpeedChangedEventArgs>(Model_SpeedChanged);

            _viewModel = new MainViewModel();
            TitleScreenViewModel titleScreen = new TitleScreenViewModel();
            _viewModel.SelectedViewModel = titleScreen;
            SubscribeToEvents(titleScreen);

            _settingsViewModel = new SettingsViewModel();
            _settingsViewModel.CloseSettings += new EventHandler(Settings_CloseSettings);
            _settingsViewModel.SaveSettings += new EventHandler(Settings_SaveSettings);

            _window.DataContext = _viewModel;
            _window.StateChanged += new EventHandler(Window_StateChanged);
            _window.LocationChanged += new EventHandler(Window_LocationChanged);

            _window.Show();
        }

        #endregion

        #region ViewModel event handlers

        private void TitleScreen_NewGame(object sender, EventArgs e)
        {
            UnsubscribeFromEvents(sender as TitleScreenViewModel);

            NewGameViewModel nVm = new NewGameViewModel();
            _viewModel.SelectedViewModel = nVm;
            SubscribeToEvents(nVm);
        }

        private async void TitleScreen_LoadGame(object sender, EventArgs e)
        {
            try
            {
                OpenFileDialog openFileDialog = new OpenFileDialog();
                openFileDialog.Title = "Load game";
                openFileDialog.Filter = "HELLo_World savegame (*.hell) | *.hell";
                if (openFileDialog.ShowDialog() == true)
                {
                    try
                    {
                        GameViewModel gVm = new GameViewModel(_model);
                        await _model.LoadGameAsync(openFileDialog.FileName);

                        UnsubscribeFromEvents(sender as TitleScreenViewModel);
                        _viewModel.SelectedViewModel = gVm;
                        SubscribeToEvents(gVm);
                    }
                    catch (GameDataException)
                    {
                        MessageBox.Show("Error occurred during load", "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                    }
                }
            }
            catch
            {
                MessageBox.Show("Can't load game", "HELLo_World", MessageBoxButton.OK, MessageBoxImage.Error);
            }
        }

        private void TitleScreen_ExitGame(object sender, EventArgs e)
        {
            _window.Close();
        }

        private void NewGameScreen_BackToTitleScreen(object sender, EventArgs e)
        {
            UnsubscribeFromEvents(sender as NewGameViewModel);

            TitleScreenViewModel titleScreen = new TitleScreenViewModel();
            _viewModel.SelectedViewModel = titleScreen;
            SubscribeToEvents(titleScreen);
        }

        private void NewGameScreen_StartNewGame(object sender, StartNewGameEventArgs e)
        {
            UnsubscribeFromEvents(sender as NewGameViewModel);

            GameViewModel gVm = new GameViewModel(_model);
            _viewModel.SelectedViewModel = gVm;
            _model.NewGame(e.Name);
            SubscribeToEvents(gVm);
        }

        private void GameScreen_BackToTitleScreen(object sender, EventArgs e)
        {
            bool restartTimer = _timer.IsEnabled;
            bool exited = false;
            _timer.Stop();
            if (MessageBox.Show("Are you sure you want to exit?\nNot saved progress will be lost!", "HELLo_World", MessageBoxButton.YesNo, MessageBoxImage.Question) == MessageBoxResult.Yes)
            {
                exited = true;
                UnsubscribeFromEvents(sender as GameViewModel);
                (sender as GameViewModel).UnSubscribeFromModelEvents();

                TitleScreenViewModel titleScreen = new TitleScreenViewModel();
                _viewModel.SelectedViewModel = titleScreen;
                SubscribeToEvents(titleScreen);
            }
            if (restartTimer && !exited) _timer.Start();
        }

        private void GameScreen_OpenSettings(object sender, EventArgs e)
        {
            _lastSpeed = _model.GameSpeed;
            _model.GameSpeed = Speed.PAUSE;

            _settingsWindow = new SettingsWindow();
            _settingsWindow.DataContext = _settingsViewModel;
            _settingsViewModel.TicketValue = _model.GameMap.Ticket;
            _settingsViewModel.MaintainerValue = _model.NumberOfRepairmen;

            _settingsWindow.Show();
        }

        private async void GameScreen_SaveGame(object sender, EventArgs e)
        {
            _lastSpeed = _model.GameSpeed;
            _model.GameSpeed = Speed.PAUSE;

            try
            {
                SaveFileDialog saveFileDialog = new SaveFileDialog();
                saveFileDialog.Title = "Save game";
                saveFileDialog.Filter = "HELLo_World savegame (*.hell) | *.hell";
                if (saveFileDialog.ShowDialog() == true)
                {
                    try
                    {
                        await _model.SaveGameAsync(saveFileDialog.FileName);
                    }
                    catch (GameDataException)
                    {
                        MessageBox.Show("Error occurred during save", "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                    }
                }
            }
            catch
            {
                MessageBox.Show("Can't save game", "HELLo_World", MessageBoxButton.OK, MessageBoxImage.Error);
            }

            _model.GameSpeed = _lastSpeed;
        }

        private void Settings_CloseSettings(object sender, EventArgs e)
        {
            _settingsWindow.Close();

            _model.GameSpeed = _lastSpeed;
        }

        private void Settings_SaveSettings(object sender, EventArgs e)
        {
            if (_model.GameMap.Ticket != _settingsViewModel.TicketValue) _model.SetTicket(_settingsViewModel.TicketValue);
            if (_model.NumberOfRepairmen != _settingsViewModel.MaintainerValue) _model.NumberOfRepairmen = _settingsViewModel.MaintainerValue;

            _settingsWindow.Close();

            _model.GameSpeed = _lastSpeed;
        }

        private void SubscribeToEvents(ViewModelBase vm)
        {
            if (vm is TitleScreenViewModel)
            {
                TitleScreenViewModel tVm = vm as TitleScreenViewModel;

                tVm.NewGame += new EventHandler(TitleScreen_NewGame);
                tVm.LoadGame += new EventHandler(TitleScreen_LoadGame);
                tVm.ExitGame += new EventHandler(TitleScreen_ExitGame);
            }
            else if (vm is NewGameViewModel)
            {
                NewGameViewModel nVm = vm as NewGameViewModel;

                nVm.BackToTitleScreen += new EventHandler(NewGameScreen_BackToTitleScreen);
                nVm.StartNewGame += new EventHandler<StartNewGameEventArgs>(NewGameScreen_StartNewGame);
            }
            else if (vm is GameViewModel)
            {
                GameViewModel gVm = vm as GameViewModel;

                gVm.BackToTitleScreen += new EventHandler(GameScreen_BackToTitleScreen);
                gVm.OpenSettings += new EventHandler(GameScreen_OpenSettings);
                gVm.SaveGame += new EventHandler(GameScreen_SaveGame);
            }
        }

        private void UnsubscribeFromEvents(ViewModelBase vm)
        {
            if (vm is TitleScreenViewModel)
            {
                TitleScreenViewModel tVm = vm as TitleScreenViewModel;

                tVm.NewGame -= TitleScreen_NewGame;
                tVm.LoadGame -= TitleScreen_LoadGame;
                tVm.ExitGame -= TitleScreen_ExitGame;
            }
            else if (vm is NewGameViewModel)
            {
                NewGameViewModel nVm = vm as NewGameViewModel;

                nVm.BackToTitleScreen -= NewGameScreen_BackToTitleScreen;
                nVm.StartNewGame -= NewGameScreen_StartNewGame;
            }
            else if (vm is GameViewModel)
            {
                GameViewModel gVm = vm as GameViewModel;

                gVm.BackToTitleScreen -= GameScreen_BackToTitleScreen;
                gVm.OpenSettings -= GameScreen_OpenSettings;
                gVm.SaveGame -= GameScreen_SaveGame;
            }
        }

        #endregion

        #region Window event handlers

        private void Window_LocationChanged(object sender, EventArgs e)
        {
            if (_viewModel.SelectedViewModel is GameViewModel)
            {
                GameViewModel gVm = _viewModel.SelectedViewModel as GameViewModel;
                if (gVm.ErrorPopup)
                {
                    var before = gVm.ErrorX;
                    gVm.ErrorX++;
                    gVm.ErrorX = before;
                }

                if (gVm.BuildingInfoPopup)
                {
                    var before = gVm.BuildingInfoX;
                    gVm.BuildingInfoX++;
                    gVm.BuildingInfoX = before;
                }

                if (gVm.VisitorInfoPopup)
                {
                    var before = gVm.VisitorInfoX;
                    gVm.VisitorInfoX++;
                    gVm.VisitorInfoX = before;
                }
            }
        }

        private void Window_StateChanged(object sender, EventArgs e)
        {
            if (_viewModel.SelectedViewModel is GameViewModel)
            {
                GameViewModel gVm = _viewModel.SelectedViewModel as GameViewModel;
                switch (_window.WindowState)
                {
                    case WindowState.Minimized:
                        gVm.BuildingInfoPopup = false;
                        gVm.VisitorInfoPopup = false;
                        gVm.ErrorPopup = false;
                        break;
                }
            }
        }

        #endregion

        #region Model event handler

        private void Model_SpeedChanged(object sender, SpeedChangedEventArgs e)
        {
            switch (e.GameSpeed)
            {
                case Speed.PAUSE:
                    _timer.Stop();
                    break;
                case Speed.SLOW:
                    _timer.Interval = TimeSpan.FromSeconds(0.4);
                    _timer.Start();
                    break;
                case Speed.NORMAL:
                    _timer.Interval = TimeSpan.FromSeconds(0.2);
                    _timer.Start();
                    break;
                case Speed.FAST:
                    _timer.Interval = TimeSpan.FromSeconds(0.1);
                    _timer.Start();
                    break;
                default:
                    break;
            }
        }

        #endregion

        #region Timer event handler

        private void Timer_Tick(object sender, EventArgs e)
        {
            _model.AdvanceTime();
        }

        #endregion
    }
}
